#ifndef LINSOLVE_H
#define LINSOLVE_H

/***************************************************************************/
/*                  UTILITAIRES MLL D'ALGEBRE LINEAIRE                     */
/*-------------------------------------------------------------------------*/
/*  D'une maniere generale, les vecteurs, tableaux, matrices sont passes   */
/*  en argument par un pointeur sur un double pointant sur le 1er element. */
/*  Ensuite, les elements des matrices sont supposes ranges en colonne.    */
/*  Par exemple dans la matrice 2 x 3 (2 lignes, 3 colonnes) suivante      */
/*                | a11 a12 a13 |                                          */
/*                | a21 a22 a23 |                                          */
/*  l'argument passe est un pointeur sur a11, et les elements sont         */
/*  supposes etre ranges en memoire dans l'ordre a11 a21 a12 a22 a13 a23   */
/*  c'est-a-dire comme en FORTRAN.                                         */
/*  Dans le cas ou l'usager utilise des tableaux bidimensionnels du type : */
/*  double a[n1][n2] ;                                                     */
/*  il devra considerer que n1 est l'indice de colonne et n2 l'indice de   */
/*  lignes pour etre en conformite avec nos routines.                      */
/*  L'argument a passer aux routines est dans ce cas, au choix :           */
/*       *a = a[0] = &a[0][0]                                              */
/*-------------------------------------------------------------------------*/
/*  A l'exception des routines inspirees de Linpack, pour lesquelles j'ai  */
/*  conserve l'ordre des arguments des routines FORTRAN, je place toujours */
/*  en tete l'argument de sortie de la routine. Exemple produit matriciel  */
/*  A = B * C --> matmat (A, B, C, ...)                                    */
/*-------------------------------------------------------------------------*/
/*  Les dimensions des matrices sont passes dans l'ordre nombre de lignes, */
/*  puis nombre de colonnes. Si plusieurs matrices sont en jeux, les       */
/*  premiers indices sont ceux de la matrice principale (par exemple dans  */
/*  Amp = Bmn*Cnp les indices sont dans l'ordre m, n, p.                   */
/*-------------------------------------------------------------------------*/
#define MAXIT 31                    /* Nb limite d'iterations pour la SVD  */
#include "spn.h" /* Pour acceder a : iamaxvec,dotvec,swapvec,longvec,
                                     tmatmat,matmat,cpyvec,tAAmat et trmat */

/*-------------------------------------------------------------------------*/
/* Resolution systeme lin. symetrique par CHOLESKI avec/sans racine carree */
/*-------------------------------------------------------------------------*/
double solvchud (double *xy, double *A, int n, int ns) ;
double solvchsq (double *xy, double *A, int n, int ns) ;
/*-------------------------------------------------------------------------*/
/*  !!!!!!!!!!    ATTENTION la matrice A est detruite       !!!!!!!!!!
		       -1
	       [X] = [A]  *  [Y]  avec [A] matrice symetrique n x n
Remarque: Seule la partie superieure de A est utilisee.
Entree :  n        Ordre du systeme
	  ns       Nombre de seconds membres
	  xy(n,ns) Valeurs des seconds membres
	  A(n,n)   Matrice symetrique du systeme
Sortie:   xy(n,ns) Solution du systeme.
	Retour : dterminant de A.
	Si 0 resultat incorrect (eventuellement echec alloc dans solvchsq).
	Si < 0 rsultat de solvchsq incertain y compris le signe du determinant.
/*-------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
/*      Resolution systeme lineaire n x n par decomposition LU              */
/*--------------------------------------------------------------------------*/
double solvlu (double *xy, double *A, int n, int ns, int Atrp) ;
/*--------------------------------------------------------------------------*/
/*
     !!!!!!    ATTENTION la matrice A est detruite       !!!!!!
  Resolution de Y = A  * X :   X = inv(A)  * Y    si Atrp = 0
	         de Y = At * X :   X = inv(At) * Y    si Atrp = 1
  A  : matrice n x n
  xy : ns vecteurs de dimension n
En entree xy contient les ns vecteurs y et en sortie les ns vecteurs x
		Renvoi le determinant
 Si Det == 0, rsolution impossible (peut tre echec allocation ?)
----------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/*         Resolution systeme lineaire n x p par decomposition QR           */
/*--------------------------------------------------------------------------*/
int solvqr (double *xy, double *A, int n, int p, int ns, double eps) ;
/*----------------------------------------------------------------------------
   !!!!!!    ATTENTION la matrice A est detruite       !!!!!!
    Calcule la solution moindre-carres du systeme y = A * x par QR
    qui peut etre sur ou sous determine.
    L'usager doit fournir une epsilon a partir duquel les colonnes de A
    ne sont plus utilisees pour calculer la solution. Les composantes de
    x correspondantes sont mises a zero.
Entree : 
     A(n,p)    matrice n lignes, p colonnes
	 xy(n,ns)  ns vecteurs a decomposer sur les colonnes de A
	 eps       seuil relatif de singularit
Sortie : 
     xy(p, ns) composantes des ns vecteurs solutions
	 Renvoi le rang de la matrice
	        -1 si echec allocation
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/*        Resolution systeme lineaire n x p par SVD                         */
/*--------------------------------------------------------------------------*/
int solvsvd (double *xy, double *a,int n,int p,int ns, double eps);
/*---------------------------------------------------------------------------
    !!!!!!!!!!    ATTENTION la matrice a est detruite       !!!!!!!!!!
    Calcule la solution moindre-carres de norme minimale du
    systeme y = a * x qui peut etre sur ou sous determine.
Entree :
    a(n,p)    matrice n lignes, p colonnes
    xy(n,ns)  ns vecteurs a decomposer sur les colonnes de a
    eps       seuil relatif de singularit
Sortie :
    xy(p, ns) composantes des ns vecteurs solutions
	Renvoi le rang de la matrice ou 
		-1 si echec allocation
		-2 si echec factorisation
---------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/
/*     Exemple de resolution systeme lineaire quelconque par la
/*     presqu'inverse avec CHOLESKI                                        */
/*-------------------------------------------------------------------------*/
double solvpi (double *x, double *a, double *y, int n, int p, int ns, double t);
/*---------------------------------------------------------------------------
    !!!!!!!!!!    ATTENTION la matrice a est detruite       !!!!!!!!!!
					   2         2
    Calcule la solution minimisant ||y-ax||  + t||x||
Entree :
    a(n,p)    matrice n lignes, p colonnes
    y(n,ns)   ns vecteurs a decomposer sur les colonnes de a
Sortie :
    x(p, ns)  composantes des ns vecteurs solutions
    Retour !=0.  dterminant de A'*A
	       0. resultat incorrect (echec alloc eventuellement)
----------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*/
/*         Rang d'une matrice par decomposition QR                         */
/*-------------------------------------------------------------------------*/
int rankqr (double *x, int n, int p, double eps);
/*---------------------------------------------------------------------------

    !!!!!!!!!!    ATTENTION la matrice x est detruite       !!!!!!!!!!
    Calcule le rang de la matrice x
    L'usager doit fournir une epsilon relatif a partir duquel on considere
	que les colonnes sont liees.

Entree :
    X(n,p)     matrice n lignes, p colonnes
    eps        seuil relatif de singularit
Sortie :
    Renvoi le rang r de X
	       -1 si echec allocation
---------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*/
/*         Noyau d'une matrice par decomposition QR                        */
/*-------------------------------------------------------------------------*/
int nullqr (double *noy, double *x, int n, int p, double eps);
/*---------------------------------------------------------------------------

    !!!!!!!!!!    ATTENTION la matrice x est detruite       !!!!!!!!!!
    Calcule le noyau de la matrice x
    L'usager doit fournir une epsilon relatif a partir duquel on considere
	que les colonnes sont liees.

Entree :
    X(n,p)     matrice n lignes, p colonnes
    eps        seuil relatif de singularit
Sortie :
    noy(n, n-r) Noyau orthonorme  gauche de X tel que N'*X = 0
    Renvoi le rang r de X
	       -1 si echec allocation
---------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/
/*                        Decomposition QR                                 */
/*-------------------------------------------------------------------------*/
int mat2QR (double *Q, double *AR, int n, int p);
/*---------------------------------------------------------------------------

    Calcule la decomposition QR de la matrice A
	            A = QR

Entree :
    XR(n,p)    matrice A n lignes, p colonnes
Sortie :
    Q(n, n)  matrice orthogonale Q
	XR(n, p)  matrice R n lignes, p colonnes
	Renvoi 0 normalement
		-1 si echec allocation
---------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/
int mat2QRE (double *Q, double *AR, int *jpvt, int n, int p, double eps);
/*---------------------------------------------------------------------------

    Calcule la decomposition QRE de la matrice A
					A = QRE'

Entree :
    XR(n,p)    matrice A n lignes, p colonnes
Sortie :
    Q(n, n)  matrice orthogonale Q
	XR(n, p)  matrice R n lignes, p colonnes
	jpvt[p]   permutation des colonnes QR qui permet de reconstituer A
	Renvoi le rang ou -1 si echec allocation.
	Pour retrouver A, il faut faire Q*R, puis appeler
	permcol(QR, jpvt, nlin, ncol);
---------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/
void permcol(double *A, int *jpvt, int nlin, int ncol);
/*-------------------------------------------------------------------------*/
/*  Permutation des colonnes d'une matrice A(nlin,ncol) selon les          */
/*  indications du tableau de permutation jpvt (non modifie)               */
/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/
void permlin(double *A, int *jpvt, int nlin, int ncol);
/*-------------------------------------------------------------------------*/
/*  Permutation des lignes d'une matrice A(nlin,ncol) selon les            */
/*  indications du tableau de permutation jpvt (non modifie)               */
/*-------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*/
/*                  Noyau d'une matrice par SVD                            */
/*-------------------------------------------------------------------------*/
int nullsvd (double *Noy, double *a, int n, int p, double eps, int fleft) ;
/*---------------------------------------------------------------------------
    !!!!!!!!!!    ATTENTION la matrice x est detruite       !!!!!!!!!!
    Calcule le noyau  droite ou  gauche (transpose) de la matrice x
    L'usager doit fournir une epsilon relatif a partir duquel on considere
	que les colonnes sont liees.

Entree :
    X(n,p)     matrice n lignes, p colonnes
    eps        seuil relatif de singularit
	fleft	   0 : calcul du noyau  droite, sinon noyau a gauche transpose.
Sortie :
    noy(p, p-r) Noyau orthonorme  droite de X si fleft = 0
    noy(n, n-r) Noyau orthonorme  gauche de X si fleft != 0
    Renvoi le rang r de X ou
		-1 si echec allocation
		-2 echec factorisation
---------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/
int mat2SVD (double *u,double *S,double *v,double *a,int n,int p,double eps);
/*---------------------------------------------------------------------------
    !!!!!!!!!!    ATTENTION la matrice a est detruite       !!!!!!!!!!
    Calcule la solution moindre-carres de norme minimale du
    systeme y = a * x qui peut etre sur ou sous determine.
Entree :
    a(n,p)    matrice n lignes, p colonnes
    eps       seuil relatif de singularit
Sortie :
    u(n, n)   
    s(k)	  k = min(n,p)
    v(p, p)  
	Renvoi le rang de la matrice ou 
		-1 si echec allocation
		-2 si echec factorisation
---------------------------------------------------------------------------*/


/* Dterminant d'une matrice par dcomposition LU
   !!!!!!!  ATTENTION LA MATRICE A EST DETRUITE   !!!!!!!!!!!!   */
double detmat(double *a, int n);
/* Si Det == 0, il y a peut tre eu echec allocation             */

/* Dterminant et Inversion d'une matrice (par dcomposition LU) */
double invmat(double *a, int n) ;
/* Si Det == 0, il y a peut tre eu echec allocation             */

/*---------------------------------------------------------------------------
 Pseudo-Inverse (SV) de la matrice a(n,p) : 
	 0 : OK
	-1 : echec allocation
	-2 : echec resolution 

 Attention a est detruite 
---------------------------------------------------------------------------*/
int pinvmat(double *x, double *a, int n, int p);


#endif
